/*
This program is presented in several fragments in the doc page.
Every fragment is in its own file; this file simply combines them.
*/

#include "make_circulant.cpp.preamble"
#include "make_circulant.cpp.traits"
#include "make_circulant.cpp.expression"
#include "make_circulant.cpp.evaluator"
#include "make_circulant.cpp.entry"
#include "make_circulant.cpp.main"
