MatrixXd A = MatrixXd::Random(3, 3);
MatrixXd B = MatrixXd::Random(3, 2);
cout << "Here is the invertible matrix A:" << endl << A << endl;
cout << "Here is the matrix B:" << endl << B << endl;
MatrixXd X = A.lu().solve(B);
cout << "Here is the (unique) solution X to the equation AX=B:" << endl << X << endl;
cout << "Relative error: " << (A * X - B).norm() / B.norm() << endl;
